//	NEBApp.swift
//
//	© 2025 by Jeff Weeks
//	See TermsOfUse.txt

import SwiftUI
import CompositorServices

//	Note:  If you go to
//
//		Target > Build Settings > Swift Compiler - Custom Flags > Other Swift Flags > Debug
//
//	and add the lines
//
//		-Xfrontend
//		-warn-long-function-bodies=100
//		-Xfrontend
//		-warn-long-expression-type-checking=5
//
//	the compiler will measure the time required for type-checking individual lines of code
//	and for compiling individual functions, and will issue warnings for slow ones.
//	Unfortunately there are many lines of code that are slow to type-check.
//	Note: Accessing a component of a vector v as v.x is faster to type check
//	than accessing it as v[0].

//#warning("makeScreenshots is enabled -- disable for public release")
let makeScreenshots: Bool = false

//	At a lecture or other public event, having each person
//	do the eye calibration procedure wastes a huge amount of time.
//	By making the game selection buttons huge, people may be able
//	to use (somebody else's!) pre-stored calibration and play
//	with mis-calibrated vision. In case of problems, I can
//	quadruple-press the Top Button and work the Digital Crown
//	for them to do the hand & eye setup without having to go
//	through all the other hassle of Guest Mode.
//
//	When people use the app only briefly at public events,
//	we'll also want to strengthen the Guardian Angel,
//	weaken the computer player, and make the table a little higher.
//
//		Caution: We want to leave all settings fully adjustable
//		for real-time flexibility. So setting different default values
//		doesn't quite guarantee that the actual values won't get
//		unadvertantly altered before the lecture begins.
//		I've left myself a reminder in my lecture notes (in Keynote)
//		to double-check all the desired values at some point
//		before the talk (they persist in User Prefs).
//
//#warning("Be sure forTalks is false when submitting to the App Store.")
let forTalks = false


@main
struct NEBApp: App {

	@AppStorage("Player handedness") var playerHandedness: PlayerHandedness = .rightHanded

	//	How well should the computer play?
	@AppStorage("Difficulty") var computerPlayerDifficulty: Double	//	∈ 0.0 ... 1.0
									 = forTalks ? 0.25 : 0.5
	
	//	Especially at public events where each player might have only
	//	a couple minutes to try the app, novices might appreciate
	//	a helping hand to make their shots more successful.
	@AppStorage("Guardian angel factor") var guardianAngelFactor: Double	//	∈ 0.0 ... 1.0
									= forTalks ? 1.0 : 0.75
	
	@State var geometryType: GeometryType = .euclidean

	@AppStorage("Table height") var tableHeight: Double = defaultTableHeight	//	in meters

	//	keepTableLevel is true by default because it lets
	//	the player hold the cue stick at a more comfortable angle.
	//
	//	Players may disable keepTableLevel
	//
	//	  - to experience fully authentic head-tracking
	//		  in non-Euclidean geometry or
	//
	//	  - to keep the north pole (where several walls meet)
	//		  out of the player's way.
	//
	@AppStorage("Keep table level") var keepTableLevel: Bool = true

	//	Should the renderer adjust the distances so that
	//	a Euclidean born-and-raised tourist will see each object
	//	at the same distance that a native-born inhabitant
	//	of the non-Euclidean space would see it?
	//
	//	For a full discussion of tourist view vs. native view, see
	//
	//		Jeff Weeks,
	//		Non-Euclidean Billiards in VR, Proceedings of Bridges 2020:
	//		Mathematics, Music, Art, Architecture, Culture (2020) 1–8.
	//
	//		https://archive.bridgesmathart.org/2020/bridges2020-1.pdf
	//
	//	and
	//
	//		Jeff Weeks,
	//		Body coherence in curved-space virtual reality games,
	//		Computers & Graphics 97 (2021) 28–41.
	//
	//		  https://www.sciencedirect.com/science/article/pii/S0097849321000443
	//		or
	//		  https://arxiv.org/abs/2011.00510
	//
	//	2024-10-03  Now that we no longer anchor the player's hands
	//	to a physical table, but instead anchor them to the player's head
	//	and map them into the space using the "exponential map",
	//	it makes sense to set adjustPerceivedDistances to true by default,
	//	so the player will see his/her hands exactly where s/he feels them to be.
	//
	//	2025-01-12  We now automatically keep adjustPerceivedDistances
	//	enabled when the player is in spherical geometry.
	//
	@AppStorage("Adjust perceived hyperbolic distances") var adjustPerceivedDistances: Bool = true

	//	Apple's .upperLimbVisibility() page
	//
	//		https://developer.apple.com/documentation/swiftui/view/upperlimbvisibility(_:)
	//
	//	says
	//
	//		Note that this modifier only sets a preference and ultimately
	//		the system will decide if it will honor it or not. The system
	//		may be unable to honor the preference if the immersive space
	//		is currently not visible.
	//
	//	My first reaction was, how could the immersive space already be visible,
	//	if we supply the .upperLimbVisibility() modifier as part of
	//	our request to create it? The resolution of this paradox is that
	//	visionOS uses the (transient!) SwiftUI ImmersiveSpace struct
	//	to create an immersive space in some private internal format.
	//	If we specify the upper-limb visibility as
	//
	//		.upperLimbVisibility(hideHands ? .hidden : .visible)
	//
	//	we can switch from visible hands to hidden hands at some time
	//	after the immersive view is already visible. In particular,
	//	changing the parameter passed to .upperLimbVisibility()
	//	doesn't cause visionOS to re-create the immersive view.
	//	It keeps the immersive view that's already visible and
	//	just changes the upper-limb visibility. Whew!
	@State var hideHands: Bool = false

	@Environment(\.openWindow) var openWindow
	@Environment(\.dismissWindow) var dismissWindow


	init() {

		//	Let the NEBSoundPlayer load its sounds
		//	well before the player takes his/her first shot.
		Task { await NEBSoundPlayer.shared.hello() }
	}

	var body: some Scene {

		WindowGroup(id: "Control panel", for: Int.self) { _ in
		
			NEBControlPanel(
				playerHandedness: $playerHandedness,
				computerPlayerDifficulty: $computerPlayerDifficulty,
				guardianAngelFactor: $guardianAngelFactor,
				geometryType: $geometryType,
				tableHeight: $tableHeight,
				keepTableLevel: $keepTableLevel,
				adjustPerceivedDistances: $adjustPerceivedDistances,
				hideHands: $hideHands)
			
		} defaultValue: { 0 }	//	assigns value 0 to the initial window
		.defaultSize(
			width: 256.0,
			height: forTalks ? 1280.0 : 352.0)

		ImmersiveSpace(id: "Billiards immersive space") {

			CompositorLayer(configuration: GGVRCompositorLayerConfiguration()
			) { layerRenderer in
			
				let theOpenControlPanelClosure = {  @Sendable () -> Void in
					Task { @MainActor in
						openWindow(id: "Control panel", value: 0)
					}
				}

#warning("layerRenderer isn't sendable, so when Strict Concurrency Checking is set to 'complete', the compiler complains here. Xcode's template code has this same problem.")
				Task(executorPreference: GGVRRendererTaskExecutor.shared) {
				
					guard let theRenderer = await NEBRenderer(
							layerRenderer: layerRenderer,
							playerHandedness: playerHandedness,
							computerPlayerDifficulty: computerPlayerDifficulty,
							guardianAngelFactor: guardianAngelFactor,
							keepTableLevel: keepTableLevel,
							adjustPerceivedDistances: adjustPerceivedDistances,
							geometryType: geometryType,
							tableHeight: tableHeight,
							openControlPanel: theOpenControlPanelClosure)
					else {
						assertionFailure("Failed to create theRenderer")
						return
					}

					await theRenderer.startWorldAndHandTracking()
					await theRenderer.runRenderLoop()
					
					//	Note: runRenderLoop() contains no 'await' calls
					//	and runs without interruption, which is just
					//	what we want for a render loop.
					//	But be aware that if some other Task
					//	tries to call one of the renderer's methods,
					//	the call will never be received.
				}
			}
		}
		.immersionStyle(selection: .constant(.full), in: .full)
		.upperLimbVisibility(hideHands ? .hidden : .visible)
		.persistentSystemOverlays(.hidden)
			//	Asks visionOS not to interpret palm-up as a request for a home button.
			//	visionOS may, at whim, choose not to comply.
	}
}
