//	NEBRenderer.swift
//
//	© 2025 by Jeff Weeks
//	See TermsOfUse.txt

import CompositorServices
import MetalKit	//	for MTKTextureLoader


let groundingShadowMeshRefinement: Int = 32
let pocketMeshRefinement: Int = 64


actor NEBRenderer {

	let layerRenderer: LayerRenderer

	let playerHandedness: PlayerHandedness
	let computerPlayerDifficulty: Double	//	∈ 0.0 ... 1.0
	let adjustPerceivedDistances: Bool
	let tableHeight: Double
	
	var openControlPanel: @Sendable () -> Void

	//	We'll access the gameState only from our render loop,
	//	so it should be thread-safe even with no locks.
	var gameState: NEBGameState
	
	var previousPresentationTime: TimeInterval?

	let arKitSession: ARKitSession
	let worldTrackingProvider: WorldTrackingProvider
	let handTrackingProvider: HandTrackingProvider

	//	With plain unmodified headset tracking,
	//	each small Euclidean motion of the physical headset
	//	gets mapped to the corresponding small motion
	//	of the player's head in the (often non-Euclidean)
	//	virtual space. While this provides a fully
	//	authentic simulation of the non-Euclidean space,
	//	it makes it harder for the player to hold
	//	the cue stick, because the table slants
	//	upward or downward from its edge.
	//
	//	To make the game easier to play, we offer
	//	the player a "Keep table level" option.
	//	When enabled, we track only the player's
	//	orthogonal projection (a.k.a. "orthogonal shadow")
	//	onto the plane of the virtual table, instead of
	//	tracking their virtual pose in the full 3D space.
	//	We then give the player's virtual head the same
	//	rotation and height above the orthogonal shadow
	//	that the player's physical head has
	//	in the physical Euclidean space.
	//
	let keepTableLevel: Bool
	var physicalHeadsetPose: simd_double4x4?	//	in physical world coordinates
	var virtualHeadsetShadow: simd_double4x4?	//	in XZW table coordinates (Y = 0)
	var virtualHeadsetPose: simd_double4x4?		//	in full table coordinates

	let computerPlayer: NEBComputerPlayer
	
	//	I'm slightly uncomfortable using implicitly unwrapped optionals
	//	-- I hadn't intended to use them at all, for run-time safety --
	//	but in this case I really would rather not have to get
	//	all the Metal objects set up directly in init(),
	//	without being able to make any function calls.
	//	So let's go ahead and use implicitly unwrapped optionals here.
	
	var device: MTLDevice!
	var commandQueue: MTLCommandQueue!
	
	var billiardsTablePipelineState: MTLRenderPipelineState!
	var groundingShadowPipelineState: MTLRenderPipelineState!
	var pocketPipelineState: MTLRenderPipelineState!
	var billiardBallPipelineState: MTLRenderPipelineState!
	var cueStickPipelineState: MTLRenderPipelineState!
	var handPipelineState: MTLRenderPipelineState!
	var checkerboardTilingPipelineState: MTLRenderPipelineState!
	var messagePipelineState: MTLRenderPipelineState!

	var enabledDepthStencilState: MTLDepthStencilState!
	var groundingShadowDepthStencilState: MTLDepthStencilState!
	var pocketDepthStencilState: MTLDepthStencilState!
	var messageDepthStencilState: MTLDepthStencilState!
	
	var tableMesh: NEBMesh!
	var groundingShadowMesh: NEBMesh!
	var pocketMesh: NEBMesh!
	var ballMesh: NEBMesh!
	var cueStickMesh: NEBMesh!
	var handFacets: MTLBuffer!
	
	var checkerboardPoses: MTLBuffer!
	var checkerboardCellWalls: [CellWall] = []
	var checkerboardCellWallsReducedLOD: [CellWall] = []
	
	var messageMesh: MTLBuffer!

	var ballTextures: [MTLTexture] = []
	var cueStickTexture: MTLTexture!
	var	messageTexture: MTLTexture?
	var messageOpacity: Float16?
	var messageFadeTimer: Double?
	
	var ballSampler: MTLSamplerState!
	var cueStickSampler: MTLSamplerState!
	var messageSampler: MTLSamplerState!

// MARK: -
// MARK: Init

	init?(
		layerRenderer: LayerRenderer,
		playerHandedness: PlayerHandedness,
		computerPlayerDifficulty: Double,	//	∈ 0.0 ... 1.0
		guardianAngelFactor: Double,		//	∈ 0.0 ... 1.0
		keepTableLevel: Bool,
		adjustPerceivedDistances: Bool,
		geometryType: GeometryType,
		tableHeight: Double,
		openControlPanel: @Sendable @escaping () -> Void
	) async {

		self.layerRenderer = layerRenderer
		self.playerHandedness = playerHandedness
		self.computerPlayerDifficulty = computerPlayerDifficulty
		self.keepTableLevel = keepTableLevel
		self.tableHeight = tableHeight

		//	When rendering in spherical space, we insist that
		//	adjustPerceivedDistances always be enabled, because
		//	if it weren't, the player would need to look walleyed
		//	when focusing on objects in the southern hemisphere.
		//	Such walleyed vision would be 100% geometrically correct,
		//	but would be unpleasant and would risk causing migraines or worse.
		//
		self.adjustPerceivedDistances
			= adjustPerceivedDistances || geometryType == .spherical

		let theSpaceGeometry = NEBSpaceGeometry(geometryType: geometryType)

		self.gameState = NEBGameState(
							spaceGeometry: theSpaceGeometry,
							guardianAngelFactor: guardianAngelFactor)

		self.computerPlayer = NEBComputerPlayer()

		self.openControlPanel = openControlPanel
		
		self.arKitSession = ARKitSession()
		self.worldTrackingProvider =  WorldTrackingProvider()
		self.handTrackingProvider = HandTrackingProvider()

		if !setUpMetalObjects() {
			return nil
		}
	}
	
	deinit {
		openControlPanel()
	}

// MARK: -
// MARK: World and hand tracking

	func startWorldAndHandTracking() async {
	
		do {
#if targetEnvironment(simulator)
			//	The simulator doesn't support hand tracking.
			try await arKitSession.run([worldTrackingProvider])
#else
			try await arKitSession.run([worldTrackingProvider, handTrackingProvider])
#endif
		} catch {
			preconditionFailure("ARKitSession error: \(error)")
		}
	}


// MARK: -
// MARK: Set up Metal objects

	func setUpMetalObjects(
	) -> Bool {
		
		device = layerRenderer.device
		
		guard let theCommandQueue = device.makeCommandQueue()
		else { return false }
		commandQueue = theCommandQueue
		
		let theCheckerboardPoseMatrices = makeCheckerboardPoseMatrices(
				geometryType: gameState.spaceGeometry.geometryType)
		
		let theFogCutoff = fogCutoff(
							for: theCheckerboardPoseMatrices,
							geometryType: gameState.spaceGeometry.geometryType)

		if  setUpPipelineStates(fogCutoff: theFogCutoff)
		 && setUpDepthStencilStates()
		 && setUpFixedBuffers(
		 		checkerboardPoseMatrices: theCheckerboardPoseMatrices)
		 && setUpTextures()
		 && setUpSamplerStates()
		{
			return true
			
		} else {
		
			assertionFailure("Unable to set up Metal resources")
			return false
		}
	}

	func setUpPipelineStates(
		fogCutoff: Double	//	in radians
	) -> Bool {

		guard let theGPUFunctionLibrary = device.makeDefaultLibrary() else {
			assertionFailure("Failed to create default GPU function library")
			return false
		}
		
		
		//	Vertex functions

		//	Apple explicitly says it's OK
		//	to modify and reuse MTLFunctionConstantValues.
		//
		let theCompileTimeConstants = MTLFunctionConstantValues()

		func setCompileTimeConstants(
			billiardBallShading: Bool = false,
			cueStickShading: Bool = false,
			antiflickering: Bool = false,
			clearDepthBuffer: Bool = false,
			instancedRendering: Bool = false,
			fog: Bool = false
		) {

			//	setConstantValue() wants reference values.
			
			var adjustPerceivedDistancesAsVar: Bool = adjustPerceivedDistances
			var geometryCodeAsVar: UInt8 = gameState.spaceGeometry.geometryType.gpuGeometryCode
			var spaceRadiusAsVar: Float32 = Float32(gameState.spaceGeometry.spaceRadius)
			var billiardBallShadingAsVar: Bool = billiardBallShading
			var cueStickShadingAsVar: Bool = cueStickShading
			var antiflickeringAsVar: Bool = antiflickering
			var clearDepthBufferAsVar: Bool = clearDepthBuffer
			var instancedRenderingAsVar: Bool = instancedRendering
			var fogAsVar: Bool = fog
			var fogCutoffAsVar = Float(fogCutoff)

			theCompileTimeConstants.setConstantValue(
				&adjustPerceivedDistancesAsVar,
				type: .bool,
				withName: "gAdjustPerceivedDistances")
			theCompileTimeConstants.setConstantValue(
				&geometryCodeAsVar,
				type: .uchar,
				withName: "gGeometryCode")
			theCompileTimeConstants.setConstantValue(
				&spaceRadiusAsVar,
				type: .float,
				withName: "gSpaceRadius")
			theCompileTimeConstants.setConstantValue(
				&billiardBallShadingAsVar,
				type: .bool,
				withName: "gBilliardBallShading")
			theCompileTimeConstants.setConstantValue(
				&cueStickShadingAsVar,
				type: .bool,
				withName: "gCueStickShading")
			theCompileTimeConstants.setConstantValue(
				&antiflickeringAsVar,
				type: .bool,
				withName: "gAntiflickering")
			theCompileTimeConstants.setConstantValue(
				&clearDepthBufferAsVar,
				type: .bool,
				withName: "gClearDepthBuffer")
			theCompileTimeConstants.setConstantValue(
				&instancedRenderingAsVar,
				type: .bool,
				withName: "gInstancedRendering")
			theCompileTimeConstants.setConstantValue(
				&fogAsVar,
				type: .bool,
				withName: "gFog")
			theCompileTimeConstants.setConstantValue(
				&fogCutoffAsVar,
				type: .float,
				withName: "gFogCutoff")
		}

		setCompileTimeConstants()
		let theGroundingShadowVertexFunction: MTLFunction
		do {
			theGroundingShadowVertexFunction = try theGPUFunctionLibrary.makeFunction(
				name: "nonEuclideanBilliardsVertexFunction",
				constantValues: theCompileTimeConstants)
		} catch {
			assertionFailure("Failed to load theGroundingShadowVertexFunction: \(error)")
			return false
		}
		theGroundingShadowVertexFunction.label = "grounding shadow vertex function"

		setCompileTimeConstants(clearDepthBuffer: true)
		let thePocketVertexFunction: MTLFunction
		do {
			thePocketVertexFunction = try theGPUFunctionLibrary.makeFunction(
				name: "nonEuclideanBilliardsVertexFunction",
				constantValues: theCompileTimeConstants)
		} catch {
			assertionFailure("Failed to load thePocketVertexFunction: \(error)")
			return false
		}
		thePocketVertexFunction.label = "pocket vertex function"

		setCompileTimeConstants(antiflickering: true)
		let theBilliardsTableVertexFunction: MTLFunction
		do {
			theBilliardsTableVertexFunction = try theGPUFunctionLibrary.makeFunction(
				name: "nonEuclideanBilliardsVertexFunction",
				constantValues: theCompileTimeConstants)
		} catch {
			assertionFailure("Failed to load theBilliardsTableVertexFunction: \(error)")
			return false
		}
		theBilliardsTableVertexFunction.label = "billiards table vertex function"

		setCompileTimeConstants(billiardBallShading: true)
		let theBallShadingVertexFunction: MTLFunction
		do {
			theBallShadingVertexFunction = try theGPUFunctionLibrary.makeFunction(
				name: "nonEuclideanBilliardsVertexFunction",
				constantValues: theCompileTimeConstants)
		} catch {
			assertionFailure("Failed to load theBallShadingVertexFunction: \(error)")
			return false
		}
		theBallShadingVertexFunction.label = "ball-shading vertex function"

		setCompileTimeConstants(cueStickShading: true)
		let theCueStickShadingVertexFunction: MTLFunction
		do {
			theCueStickShadingVertexFunction = try theGPUFunctionLibrary.makeFunction(
				name: "nonEuclideanBilliardsVertexFunction",
				constantValues: theCompileTimeConstants)
		} catch {
			assertionFailure("Failed to load theCueStickShadingVertexFunction: \(error)")
			return false
		}
		theCueStickShadingVertexFunction.label
			= "cue stick shading vertex function"

		setCompileTimeConstants()
		let theHandVertexFunction: MTLFunction
		do {
			theHandVertexFunction = try theGPUFunctionLibrary.makeFunction(
				name: "nonEuclideanBilliardsVertexFunction",
				constantValues: theCompileTimeConstants)
		} catch {
			assertionFailure("Failed to load theHandVertexFunction: \(error)")
			return false
		}
		theHandVertexFunction.label = "hand vertex function"

		//	Only the checkerboard tiling (not the table, balls,
		//	cue stick, etc.) gets fogged. Moreover, we disable
		//	even the checkerboard tiling's fog in spherical geometry
		//	so the player can see the whole lattice clearly.
		//	In any case there's no need for fog there because,
		//	unlike in Euclidean and hyperbolic geometry,
		//	the spherical tiling has no border that needs
		//	to be hidden, nor is there any problem with
		//	distinguishing overlapping walls of the same color.
		//
		let theFogFlag = gameState.spaceGeometry.geometryType == .spherical ?
							false : true
		setCompileTimeConstants(
			instancedRendering: true,
			fog: theFogFlag)
		let theCheckerboardTilingVertexFunction: MTLFunction
		do {
			theCheckerboardTilingVertexFunction = try theGPUFunctionLibrary.makeFunction(
				name: "nonEuclideanBilliardsVertexFunction",
				constantValues: theCompileTimeConstants)
		} catch {
			assertionFailure("Failed to load theCheckerboardTilingVertexFunction: \(error)")
			return false
		}
		theCheckerboardTilingVertexFunction.label = "checkerboard tiling vertex function"


		//	Fragment functions
		
		guard let theBilliardsTableFragmentFunction
			= theGPUFunctionLibrary.makeFunction(name: "billiardsTableFragmentFunction")
		else {
			assertionFailure("Failed to load billiardsTableFragmentFunction")
			return false
		}
		theBilliardsTableFragmentFunction.label = "billiards table fragment function"

		guard let theGroundingShadowFragmentFunction
			= theGPUFunctionLibrary.makeFunction(name: "groundingShadowFragmentFunction")
		else {
			assertionFailure("Failed to load groundingShadowFragmentFunction")
			return false
		}
		theGroundingShadowFragmentFunction.label = "grounding shadow fragment function"

		guard let thePocketFragmentFunction
			= theGPUFunctionLibrary.makeFunction(name: "pocketFragmentFunction")
		else {
			assertionFailure("Failed to load pocketFragmentFunction")
			return false
		}
		thePocketFragmentFunction.label = "pocket fragment function"

		guard let theBilliardBallFragmentFunction
			= theGPUFunctionLibrary.makeFunction(name: "billiardBallFragmentFunction")
		else {
			assertionFailure("Failed to load billiardBallFragmentFunction")
			return false
		}
		theBilliardBallFragmentFunction.label = "billiard ball fragment function"

		guard let theCueStickFragmentFunction
			= theGPUFunctionLibrary.makeFunction(name: "cueStickFragmentFunction")
		else {
			assertionFailure("Failed to load cueStickFragmentFunction")
			return false
		}
		theCueStickFragmentFunction.label = "cue stick fragment function"

		guard let theHandFragmentFunction
			= theGPUFunctionLibrary.makeFunction(name: "handFragmentFunction")
		else {
			assertionFailure("Failed to load theHandFragmentFunction")
			return false
		}
		theHandFragmentFunction.label = "hand fragment function"

		guard let theCheckerboardTilingFragmentFunction
			= theGPUFunctionLibrary.makeFunction(name: "checkerboardTilingFragmentFunction")
		else {
			assertionFailure("Failed to load theCheckerboardTilingFragmentFunction")
			return false
		}
		theCheckerboardTilingFragmentFunction.label = "checkerboard tiling fragment function"
		
		
		//	Vertex descriptor

		//	Create a MTLVertexDescriptor describing the vertex attributes
		//	that the GPU sends along to each vertex. Each vertex attribute
		//	may be a 1-, 2-, 3- or 4-component vector.
		
		let theVertexDescriptor = MTLVertexDescriptor()
		
			//	Say where to find each attribute.

		guard let thePosOffset = MemoryLayout.offset(of: \NEBVertex.pos) else {
			assertionFailure("Failed to find offset of NEBVertex.pos")
			return false
		}
		theVertexDescriptor.attributes[VertexAttributePosition].format = MTLVertexFormat.float4
		theVertexDescriptor.attributes[VertexAttributePosition].bufferIndex = BufferIndexVertexAttributes
		theVertexDescriptor.attributes[VertexAttributePosition].offset = thePosOffset

		guard let theTexOffset = MemoryLayout.offset(of: \NEBVertex.tex) else {
			assertionFailure("Failed to find offset of NEBVertex.tex")
			return false
		}
		theVertexDescriptor.attributes[VertexAttributeTexCoords].format = MTLVertexFormat.float3
		theVertexDescriptor.attributes[VertexAttributeTexCoords].bufferIndex = BufferIndexVertexAttributes
		theVertexDescriptor.attributes[VertexAttributeTexCoords].offset = theTexOffset

			//	Say how to step through each buffer.

		theVertexDescriptor.layouts[BufferIndexVertexAttributes].stepFunction = MTLVertexStepFunction.perVertex
		theVertexDescriptor.layouts[BufferIndexVertexAttributes].stride = MemoryLayout<NEBVertex>.stride


		//	Pipeline states.

		guard let theBilliardsTablePipelineState = ggvrPipelineState(
				layerRenderer: layerRenderer,
				vertexFunction: theBilliardsTableVertexFunction,
				fragmentFunction: theBilliardsTableFragmentFunction,
				vertexDescriptor: theVertexDescriptor,
				enableBlending: false)
		else { return false }
		billiardsTablePipelineState = theBilliardsTablePipelineState
		
		guard let theGroundingShadowPipelineState = ggvrPipelineState(
				layerRenderer: layerRenderer,
				vertexFunction: theGroundingShadowVertexFunction,
				fragmentFunction: theGroundingShadowFragmentFunction,
				vertexDescriptor: theVertexDescriptor,
				enableBlending: true)
		else { return false }
		groundingShadowPipelineState = theGroundingShadowPipelineState

		guard let thePocketPipelineState = ggvrPipelineState(
				layerRenderer: layerRenderer,
				vertexFunction: thePocketVertexFunction,
				fragmentFunction: thePocketFragmentFunction,
				vertexDescriptor: theVertexDescriptor,
				enableBlending: false)
		else { return false }
		pocketPipelineState = thePocketPipelineState
		
		guard let theBilliardBallPipelineState = ggvrPipelineState(
				layerRenderer: layerRenderer,
				vertexFunction: theBallShadingVertexFunction,
				fragmentFunction: theBilliardBallFragmentFunction,
				vertexDescriptor: theVertexDescriptor,
				enableBlending: false)
		else { return false }
		billiardBallPipelineState = theBilliardBallPipelineState

		guard let theCueStickPipelineState = ggvrPipelineState(
				layerRenderer: layerRenderer,
				vertexFunction: theCueStickShadingVertexFunction,
				fragmentFunction: theCueStickFragmentFunction,
				vertexDescriptor: theVertexDescriptor,
				enableBlending: false)
		else { return false }
		cueStickPipelineState = theCueStickPipelineState
		
		guard let theHandPipelineState = ggvrPipelineState(
				layerRenderer: layerRenderer,
				vertexFunction: theHandVertexFunction,
				fragmentFunction: theHandFragmentFunction,
				vertexDescriptor: theVertexDescriptor,
				enableBlending: false)
		else { return false }
		handPipelineState = theHandPipelineState
		
		guard let theCheckerboardTilingPipelineState = ggvrPipelineState(
				layerRenderer: layerRenderer,
				vertexFunction: theCheckerboardTilingVertexFunction,
				fragmentFunction: theCheckerboardTilingFragmentFunction,
				vertexDescriptor: theVertexDescriptor,
				enableBlending: false)
		else { return false }
		checkerboardTilingPipelineState = theCheckerboardTilingPipelineState
		
		guard let theMessagePipelineState = makeMessagePipelineState(
			gpuFunctionLibrary: theGPUFunctionLibrary,
			layerRenderer: layerRenderer)
		else { return false }
		messagePipelineState = theMessagePipelineState

		return true
	}
		
	func setUpDepthStencilStates() -> Bool {

		if let theEnabledDepthStencilState = ggvrDepthStencilState(
												device: device,
												depthCompareEnabled: true,
												depthWriteEnabled: true)
		{
			enabledDepthStencilState = theEnabledDepthStencilState
		} else {
			return false
		}

		if let theGroundingShadowDepthStencilState = ggvrDepthStencilState(
												device: device,
												depthCompareEnabled: false,
												depthWriteEnabled: false)
		{
			groundingShadowDepthStencilState = theGroundingShadowDepthStencilState
		} else {
			return false
		}

		//	The pockets will need to clear the depth buffer beneath them,
		//	so keep depthWriteEnabled = true.
		if let thePocketDepthStencilState = ggvrDepthStencilState(
												device: device,
												depthCompareEnabled: false,
												depthWriteEnabled: true)
		{
			pocketDepthStencilState = thePocketDepthStencilState
		} else {
			return false
		}

		if let theMessageDepthStencilState = makeMessageDepthStencilState(device: device)
		{
			messageDepthStencilState = theMessageDepthStencilState
		} else {
			return false
		}

		return true
	}
	
	func setUpFixedBuffers(
		checkerboardPoseMatrices: [simd_double4x4]
	) -> Bool {
	
		tableMesh = makeTableMesh(
						device: device,
						spaceGeometry: gameState.spaceGeometry)

		groundingShadowMesh = makeDiskMesh(
						device: device,
						r: gameState.spaceGeometry.ballShadowRadius,
						n: groundingShadowMeshRefinement,
						spaceGeometry: gameState.spaceGeometry)

		pocketMesh = makeDiskMesh(
						device: device,
						r: pocketRadius,
						n: pocketMeshRefinement,
						spaceGeometry: gameState.spaceGeometry)

		ballMesh = makeBallMesh(
						device: device,
						spaceGeometry: gameState.spaceGeometry)
		
		cueStickMesh = makeCueStickMesh(
						device: device,
						spaceGeometry: gameState.spaceGeometry)
		
		handFacets = ggvrMeshFacets(
						device: device,
						vertexIndices: gHandMeshFacets)

		checkerboardPoses = makeCheckerboardPoseBuffer(
						device: device,
						checkerboardPoseMatrices: checkerboardPoseMatrices)
		
		switch gameState.spaceGeometry.geometryType {

		case .spherical:

			//	adjustPerceivedDistances is always enabled
			//	in spherical geometry, so for the walls to align
			//	with the pool table we need to use exactly
			//	5 recursive subdivisions. Fortunately the performance hit
			//	should be minimal, because the octahedral tiling
			//	of the 3-sphere contains only 16 tiles and
			//	we render only 8 of them, checkboard style.
			checkerboardCellWalls = makeCheckerboardCellWalls(
				device: device,
				spaceGeometry: gameState.spaceGeometry,
				numRecursiveSubdivisions: 5,
				withFlange: false)
			
		case .euclidean:

			//	In Euclidean geometry we never need to subdivide.
			checkerboardCellWalls = makeCheckerboardCellWalls(
				device: device,
				spaceGeometry: gameState.spaceGeometry,
				numRecursiveSubdivisions: 0,
				withFlange: false)
			
		case .hyperbolic:

			if adjustPerceivedDistances {
			
				//	This is the tough case.
				//
				//	For the walls to align with the pool table
				//	we need to use exactly 5 recursive subdivisions,
				//	but we'll be rendering many thousands of dodecahedra
				//	and the GPU load would be unacceptable.
				//
				//	So we use a level-of-detail ("LOD") approach instead.
				//	encodeCommands() uses the high LOD checkerboardCellWalls
				//	to draw the central dodecahedron and its 30 immediate
				//	neighbors and then draws all remaining instances
				//	using checkerboardCellWallsReducedLOD. To avoid
				//	visible gaps where a reduced LOD dodecahedron meets
				//	a high LOD one, we give each wall of the low LOD
				//	dodecahedron a small flange so that it slightly overlaps
				//	its neighbors.

				checkerboardCellWalls = makeCheckerboardCellWalls(
					device: device,
					spaceGeometry: gameState.spaceGeometry,
					numRecursiveSubdivisions: 5,
					withFlange: false)
				
				checkerboardCellWallsReducedLOD = makeCheckerboardCellWalls(
					device: device,
					spaceGeometry: gameState.spaceGeometry,
					numRecursiveSubdivisions: 0,
					withFlange: true)
			
			} else {	//	adjustPerceivedDistances = false

				//	When adjustPerceivedDistances is false,
				//	all lines look straight to both eyes,
				//	so no subdivisions are needed.
				checkerboardCellWalls = makeCheckerboardCellWalls(
					device: device,
					spaceGeometry: gameState.spaceGeometry,
					numRecursiveSubdivisions: 0,
					withFlange: false)
			}
		}

		messageMesh = makeMessageMesh(device: device)

		return true
	}
	
	func setUpTextures() -> Bool {
	
		let theTextureLoader = MTKTextureLoader(device: device)
		let theTextureLoaderOptions = [
		
			MTKTextureLoader.Option.textureUsage
				: NSNumber(value: MTLTextureUsage.shaderRead.rawValue),
				
			MTKTextureLoader.Option.textureStorageMode
				: NSNumber(value: MTLStorageMode.`private`.rawValue),
			
			//	Mipmaps
			//
			//	Our Asset Catalog automatically generates mipmaps
			//	for all textures (at compile time? at run time? I'm not sure).
			//	To see the mipmap setting, go to NEBTextures.xcassets,
			//	open the Xcode's right-hand sidebar, and select
			//	the last tab (the Attributes inspector). Then select
			//	one of the textures, for example
			//
			//		cue stick > Universal
			//
			//	and look in the right-hand side bar to see
			//	its Mipmap Set options. Note that you have to select
			//	a specific image -- here "Universal" -- for this to work;
			//	if you'd instead just selected "cue stick" you wouldn't
			//	see the Mipmap Set options.
			//
			//	Regarding MTKTextureLoader's mipmap-related options...
			//	Apple's page
			//
			//		https://developer.apple.com/documentation/metalkit/mtktextureloader/option/1535977-allocatemipmaps
			//
			//	says that
			//
			//		[The allocateMipmaps] option applies only if the texture
			//		being loaded does not contain mipmap data.
			//		...
			//		If this key is not specified and the image being loaded
			//		contains data for mipmaps, the mipmap memory is allocated
			//		and the image data is loaded.
			//
			//	so we can omit
			//
			//		MTKTextureLoader.Option.allocateMipmaps : NSNumber(value: true),
			//
			//	Similarly, Apple's page
			//
			//		https://developer.apple.com/documentation/metalkit/mtktextureloader/1645851-newtexture
			//
			//	says that
			//
			//		"When using [newTexture()], the texture loader
			//		ignores the generateMipmaps … option."
			//
			//	Instead, the Asset Catalog provides the mipmaps.
			//	So we can omit
			//
			//		MTKTextureLoader.Option.generateMipmaps : NSNumber(value: true)
			//
		]
		
		//	Cube map coordinate conventions
		//
		//	Apple's MTLTexture documentation says which index (in 0...5)
		//	maps to which face of the cube, but doesn't say how it maps
		//	to that face. In practice the layout seems to be
		//
		//
		//		       +------+                   +---→ s
		//		       |      |                   |
		//		       |  y+  |                   ↓ t
		//		       |      |
		//		+------+------+------+------+
		//		|      |      |      |      |
		//		|  x-  |  z+  |  x+  |  z-  |
		//		|      |      |      |      |
		//		+------+------+------+------+
		//		       |      |
		//		       |  y-  |
		//		       |      |
		//		       +------+
		//
		//
		//	with each face oriented the same as in the old
		//	OpenGL ARB_texture_cube_map extension, namely
		//
		//				   largest
		//		  index     coord   pre-s  pre-t
		//		  -----    -------  -----  -----
		//			0       x > 0    -z     -y
		//			1       x < 0    +z     -y
		//			2       y > 0    +x     +z
		//			3       y < 0    +x     -z
		//			4       z > 0    +x     -y
		//			5       z < 0    -x     -y
		//
		//	with the final s and t texture coordinates given by
		//
		//		s = ( ( pre-s / largest-coord ) + 1 ) / 2
		//		t = ( ( pre-t / largest-coord ) + 1 ) / 2
		//
		//	These conventions agree with the diagram shown above.
		//
		//	Note:  An Xcode Asset Catalog's images seem to be
		//	upside down relative to the conventions just described.
		//	Most likely this is just a question of whether
		//	the image rows are being read top-down or bottom-up.
		//
		ballTextures = []
		do {
			for i in 0 ..< numBalls {
				let theBallTexture = try theTextureLoader.newTexture(
										name: "\(i) ball",
										scaleFactor: 1.0,
										bundle: nil,
										options: theTextureLoaderOptions)
				ballTextures.append(theBallTexture)
			}
		} catch {
			assertionFailure("Failed to load one or more ball textures. Error:  \(error)")
			return false
		}
	
		do {
			cueStickTexture = try theTextureLoader.newTexture(
									name: "cue stick",
									scaleFactor: 1.0,
									bundle: nil,
									options: theTextureLoaderOptions)
		} catch {
			assertionFailure("Failed to load cue stick textures. Error:  \(error)")
			return false
		}
		
		//	updateGameMessageState() triggers the creation
		//	of the messageTexture.
		messageTexture = nil
		
		return true
	}
	
	func setUpSamplerStates() -> Bool {

		//	Presumably the MTLSamplerAddressModes are irrelevant with a cube map?
		guard let theBallSampler = ggvrSamplerState(
									device: device,
									sMode: .repeat,
									tMode: .repeat)
		else {
			assertionFailure("Couldn't create theBallSampler")
			return false
		}
		ballSampler = theBallSampler
		
		//	tMode:.repeat is essential to accommodate meridional phase shift.
		guard let theCueStickSampler = ggvrSamplerState(
										device: device,
										sMode: .clampToEdge,
										tMode: .repeat)
		else {
			assertionFailure("Couldn't create cueStickSampler")
			return false
		}
		cueStickSampler = theCueStickSampler
		
		guard let theMessageSampler = makeMessageSampler(device: device)
		else {
			assertionFailure("Couldn't create cueStickSampler")
			return false
		}
		messageSampler = theMessageSampler
		
		return true
	}


// MARK: -
// MARK: Render loop

	func runRenderLoop() {

		var keepRendering = true	//	remains true until the immersive space gets closed

		while keepRendering {
		
			switch layerRenderer.state {
			
			case .paused:
				layerRenderer.waitUntilRunning()
				
			case .running:
				autoreleasepool {
					renderOneFrame()
				}
				
			case .invalidated:
				fallthrough
			@unknown default:
				keepRendering = false
			}
		}
		
		arKitSession.stop()
	}

	func renderOneFrame() {
	
		guard let theFrame = layerRenderer.queryNextFrame(),
			  let theTiming = theFrame.predictTiming()
		else {
			assertionFailure("Couldn't get theFrame and/or theTiming")
			return
		}
		
// MARK: -
// MARK: "Update" section

		//	Update anything that doesn't depend on the headset position.
		//	The startUpdate() / endUpdate() calls help
		//	Compositor Services plan its timing for future frames.

		theFrame.startUpdate()

		//	Compute the frame interval, which should be
		//	1/90 sec on a Vision Pro or 1/60 sec in the simulator.
		let theFrameInterval: Double
		let thePredictedPresentationTime = LayerRenderer.Clock.Instant.epoch.duration(
											to: theTiming.presentationTime).timeInterval
		if let thePreviousPresentationTime = previousPresentationTime {
			theFrameInterval = thePredictedPresentationTime
							 - thePreviousPresentationTime
		} else {
			theFrameInterval = 1.0/90.0	//	should occur only once, on the very first frame
		}
		previousPresentationTime = thePredictedPresentationTime	//	non-nil

		//	Did either we or our computer opponent just now take a shot?

		let theCueStick = gameState.cueStick[gameState.whoseTurn]!
		
		let theActivePlayerTookAShot
			= gameState.maybeTakeShot(
				cueStick: theCueStick,
				handTrackingInterval: theFrameInterval)

		//	The same code that lets the computer take a shot also serves
		//	to let our "computer self" play when running in the simulator,
		//	so let's enclose it in a function that can be called in both situations.
		func maybeLetComputerPlayerShoot(
			shooterStatus: PlayerStatus
		) {
			if !gameState.ballsAreRolling()
			&& !computerPlayer.shotIsInProgress()
			&& !gameState.gameIsOver {
		
				//	If the computer player has the cue ball inHand,
				//	place it on the table.
				if gameState.balls[0].status == .inHand {
					gameState.inHandCueBallPose = nil
					gameState.balls[0].placement = validCueBallPlacementNearCenter(in: gameState)
					gameState.balls[0].status = .inPlay
					Task { await NEBSoundPlayer.shared.play(.ballOnTable, volume: 1.0) }
				}

				//	Let the computer player initiate the shot.
				computerPlayer.planShot(
								in: gameState,
								shooterStatus: shooterStatus,
								difficulty: computerPlayerDifficulty,
								startTime: thePredictedPresentationTime)
			}
		}
		
		//	Is it time for our computer opponent to initiate a shot?
		if gameState.whoseTurn == .computer {
			maybeLetComputerPlayerShoot(shooterStatus: gameState.playerStatus[.computer]!)
		}

#if targetEnvironment(simulator)
		//	There's no hand-tracking in the simulator and therefore
		//	no way to take a shot. So when running in the simulator,
		//	let's let the computer play for "self".
		//	Is it time for our computer self to initiate a shot?
		if gameState.whoseTurn == .human {
			maybeLetComputerPlayerShoot(shooterStatus: gameState.playerStatus[.human]!)
		}
#endif

		//	Advance the animation for one frame interval.
		_ = gameState.update(for: theFrameInterval)
		
		theFrame.endUpdate()

// MARK: -
// MARK: Get drawable

		//	Wait until the optimal time for querying
		//	the headset and hand positions.
		LayerRenderer.Clock().wait(until: theTiming.optimalInputTime)

		//	Ask for theDrawable here, before calling startSubmission(),
		//	to avoid the risk of either not calling endSubmission(),
		//	or calling it without having submitted a frame
		//	(which would generate a run-time error).
		//
		//		Note: Once we get a drawable, we're committed
		//		to submitting a frame. Otherwise we get an error
		//		message about "more than 3 frames in flight".
		//
		guard let theDrawable = theFrame.queryDrawable()
		else {
			//	When dismissing the immersive space, we may fail
			//	to get theDrawable on our last pass through the render loop.
			//	This is OK.
			return
		}
		
// MARK: -
// MARK: "Submission" section

		//	Query the headset and hand positions
		//	and submit calls to render the frame.
		//	The startSubmission() / endSubmission() calls help
		//	Compositor Services plan its timing for future frames.

		theFrame.startSubmission()

		let thePresentationTime = LayerRenderer.Clock.Instant.epoch.duration(to:
			theDrawable.frameTiming.presentationTime).timeInterval

		if let theDeviceAnchor = worldTrackingProvider.queryDeviceAnchor(
										atTimestamp: thePresentationTime) {
										
			let thePhysicalHeadsetPoseInWorldCoords = convertFloat4x4toDouble4x4(
				theDeviceAnchor.originFromAnchorTransform)
			
			let thePreviousPhysicalHeadsetPose
				= physicalHeadsetPose ?? thePhysicalHeadsetPoseInWorldCoords

			physicalHeadsetPose	// for next time
				= thePhysicalHeadsetPoseInWorldCoords

			func virtualPoseAboveShadow(
				physicalPose: simd_double4x4
			) -> simd_double4x4 {	//	returns virtual pose
			
				var thePhysicalPoseAboveShadow = physicalPose
				thePhysicalPoseAboveShadow[3][0] = 0.0
				thePhysicalPoseAboveShadow[3][1] -= tableHeight
				thePhysicalPoseAboveShadow[3][2] = 0.0
				
				let theVirtualPoseAboveShadow = virtualPose(
					correspondingToPhysicalPose: thePhysicalPoseAboveShadow,
					spaceGeometry: gameState.spaceGeometry)
									
				return theVirtualPoseAboveShadow
			}

			if keepTableLevel {
			
				let theShadowIncrementInMeters
					= thePhysicalHeadsetPoseInWorldCoords[3]
					- thePreviousPhysicalHeadsetPose[3]

				let theShadowIncrementInVirtualRadians
					= theShadowIncrementInMeters
					/ gameState.spaceGeometry.spaceRadius

				let theVirtualShadowIncrement
					= geodesicTranslation(
						v: SIMD3<Double>(
							theShadowIncrementInVirtualRadians[0],
							0.0,	//	shadow ignores y coordinate
							theShadowIncrementInVirtualRadians[2]
						),
						geometryType: gameState.spaceGeometry.geometryType)
				
				let thePreviousShadowPose
					= virtualHeadsetShadow ?? matrix_identity_double4x4
				
				let theIncrementedShadowPose	//	right-to-left matrix action
					= thePreviousShadowPose
					* theVirtualShadowIncrement

				let theNewShadowPose = applyGramSchmidt(
					to: theIncrementedShadowPose,
					geometryType: gameState.spaceGeometry.geometryType)

				let theVirtualPoseAboveShadow = virtualPoseAboveShadow(
					physicalPose: thePhysicalHeadsetPoseInWorldCoords)
				
				let theNewVirtualPose	//	right-to-left matrix action
					= theNewShadowPose
					* theVirtualPoseAboveShadow
				
				virtualHeadsetShadow = theNewShadowPose
				virtualHeadsetPose = theNewVirtualPose
			
			} else {	//	!keepTableLevel
			
				let thePhysicalIncrement	//	right-to-left matrix action
					= thePreviousPhysicalHeadsetPose.inverse
					* thePhysicalHeadsetPoseInWorldCoords
				
				let theVirtualIncrement = virtualPose(
					correspondingToPhysicalPose: thePhysicalIncrement,
					spaceGeometry: gameState.spaceGeometry)

				let thePreviousVirtualHeadsetPose
					= virtualHeadsetPose ??
						virtualPoseAboveShadow(
							physicalPose: thePreviousPhysicalHeadsetPose)

				let theIncrementedPose
					= thePreviousVirtualHeadsetPose	//	right-to-left matrix action
					* theVirtualIncrement

				let theNewVirtualHeadsetPose = applyGramSchmidt(
					to: theIncrementedPose,
					geometryType: gameState.spaceGeometry.geometryType)

				virtualHeadsetPose = theNewVirtualHeadsetPose
			}
	
	
			//	On the one hand, Apple's web page
			//		https://developer.apple.com/documentation/compositorservices/layerrenderer/drawable/4295946-deviceanchor
			//	says
			//		At display time, Compositor Services compares your predicted
			//		device anchor with the actual position and orientation
			//		of the device at that time. If the two values don’t match,
			//		Compositor Services adjusts the pixels of your frame
			//		to align it with the hardware. If you don’t want it
			//		to make this adjustment, set this property to nil.
			//
			//	On the other hand, the documentation in the source code for
			//		LayerRenderer.Drawable.deviceAnchor
			//	says
			//		[deviceAnchor] is used by the compositor to reproject
			//		a drawable when no new drawables are available.
			//	where "reproject" means to re-present the frame but
			//	shifted by a few pixels according to where the user's head is.
			//
			//	In practice, if one sets theDrawable.deviceAnchor = nil
			//	one gets an error message
			//
			//		Presenting a drawable without a device anchor.
			//		This drawable won't be presented.
			//
			//	so it seems that the web page is wrong and
			//	the source code documentation is correct.
			//
			theDrawable.deviceAnchor = theDeviceAnchor
		
		} else {
		
			//	When the immersive space first appears,
			//	the device anchor is nil for the first frame or so.
			//	Later on, if the Vision Pro loses sight of where it is
			//	(for example, if its cameras are blocked or the room goes dark),
			//	then presumably the device anchor would again become nil.
			//	In these cases, just leave the virtualHeadsetPose where it is.
			//
			physicalHeadsetPose = nil

			theDrawable.deviceAnchor = nil
		}

		//	Hand tracking

		var theVirtualHandVertexArrays: [PlayersHand: [NEBVertex]] = [:]
		if let thePhysicalHeadsetPoseInWorldCoords = physicalHeadsetPose {

			//	latestAnchors gives a *much* smoother animation than handAnchors(at:)
			let theHandAnchors = handTrackingProvider.latestAnchors

			//	Convert the joints' poses to virtual table coordinates.
			let theJointPoses: [PlayersHand : VirtualJointPoses] = PlayersHand.allCases.reduce(into: [:]) { jointPosesDictionary, playersHand in

				let theHandAnchor: HandAnchor?
				switch playersHand {
				case .left:  theHandAnchor = theHandAnchors.leftHand
				case .right: theHandAnchor = theHandAnchors.rightHand
				}
				
				let theJointPoses
					=  VirtualJointPoses(
						for: theHandAnchor,
						spaceGeometry: gameState.spaceGeometry,
						physicalHeadsetPose: thePhysicalHeadsetPoseInWorldCoords,
						virtualHeadsetPose: virtualHeadsetPose)	//	in table coords
				
				jointPosesDictionary[playersHand] = theJointPoses
			}

			theVirtualHandVertexArrays
				= theJointPoses.mapValues() { virtualJointPoses in
			
				let theHandVertexPositions
					= handMeshVertices(virtualJointPoses: virtualJointPoses)
				
				let theHandVertices = theHandVertexPositions.map() { v in
					NEBVertex(
						pos: SIMD4<Float32>(v),
						tex: .zero)
				}
				
				return theHandVertices
			}

			let theGripHandJointPoses: VirtualJointPoses?	//	player's rear hand
			let theBridgeHandJointPoses: VirtualJointPoses?	//	player's forward hand
			switch playerHandedness {
			
			case .leftHanded:
				theGripHandJointPoses   = theJointPoses[.left]
				theBridgeHandJointPoses = theJointPoses[.right]

			case .rightHanded:
				theGripHandJointPoses   = theJointPoses[.right]
				theBridgeHandJointPoses = theJointPoses[.left]
			}

			//	Pick the two points we'd like the cue stick to pass through,
			//	in virtual table coordinates.
			let theGripPoint   = theGripHandJointPoses?.cueStickGripPoint()
			let theBridgePoint = theBridgeHandJointPoses?.cueStickBridgePoint()

			if gameState.updateGameMessageState(
				physicalHeadsetPoseInWorldCoordinates: thePhysicalHeadsetPoseInWorldCoords,
				gripPointInVirtualCoordinates: theGripPoint,
				activePlayerTookShot: theActivePlayerTookAShot) {
			
				//	The gameMessageState has changed.
				//	Update the messageTexture accordingly.
				
				let theMessageText = gameState.gameMessageState.messageText(
												hand: playerHandedness.hand,
												geometryType: gameState.spaceGeometry.geometryType,
												howGameEnded: gameState.howGameEnded)

				if !theMessageText.isEmpty {

					messageTexture = makeMessageTexture(
										device: device,
										commandQueue: commandQueue,
										inputText: theMessageText )
					messageOpacity = 1.0
					messageFadeTimer = nil
										
				} else {
				
					//	theMessageText is now empty. But rather than
					//	making the previous messageTexture disappear abruptly,
					//	let's leave it in place and fade it way gradually.
					
					messageFadeTimer = gMessageFadeDuration
				}
			}
			if var theMessageFadeTimer = messageFadeTimer {
			
				theMessageFadeTimer -= theFrameInterval
				if theMessageFadeTimer <= 0.0 {
				
					messageTexture = nil
					messageOpacity = nil
					messageFadeTimer = nil
					
				} else {
				
					//	Because human perceive brightness non-linearly,
					//	a linear fade seems too abrupt at the end.
					//	So let's try a quadratic fade instead.
					let theFractionalTime = theMessageFadeTimer / gMessageFadeDuration
					messageOpacity = Float16(theFractionalTime * theFractionalTime)
					messageFadeTimer = theMessageFadeTimer
				}
			}

	#if targetEnvironment(simulator)
			let theOwnCueStickPose = gameState.whoseTurn == .human ?
				computerPlayer.cueStickPose(atTime: thePresentationTime) : nil
	#else
			let theOwnCueStickPose
				= gameState.gameMessageState.cueStickIsAvailable ?
					getVirtualCueStickPose(
						gripPoint: theGripPoint,
						bridgePoint: theBridgePoint,
						cueBallIsInHand: gameState.balls[0].status == .inHand,
						spaceGeometry: gameState.spaceGeometry) :
					nil
	#endif
			gameState.cueStick[.human]!.movedTo(theOwnCueStickPose)

			let theComputersCueStickPose: simd_double4x4?
			switch gameState.whoseTurn {
			case .human:
				theComputersCueStickPose = nil
			case .computer:
				theComputersCueStickPose = computerPlayer.cueStickPose(atTime: thePresentationTime)
			}

			gameState.cueStick[.computer]!.movedTo(theComputersCueStickPose)
			
			if gameState.balls[0].status == .inHand {
			
				switch gameState.whoseTurn {
				
				case .human:
				
					let theCueBallCenterPoint = theGripHandJointPoses?.cueBallInHandPoint()
				
					gameState.inHandCueBallPose
						= gameState.gameMessageState.cueBallIsAvailable ?
							getVirtualInHandCueBallPose(
								cueBallCenterPoint: theCueBallCenterPoint,
								spaceGeometry: gameState.spaceGeometry) :
							nil
					
					gameState.maybePlaceCueBall()
				
				case .computer:
				
					let theHeightAboveTable = 0.35	//	in meters

					gameState.inHandCueBallPose = axisAlignedTranslation(
						axis: 1,
						distanceInRadians: theHeightAboveTable / gameState.spaceGeometry.spaceRadius,
						geometryType: gameState.spaceGeometry.geometryType)
				}
			}
		}


		guard let theCommandBuffer = commandQueue.makeCommandBuffer() else {
			preconditionFailure("Failed to make theCommandBuffer")
		}
		renderTheFrame(
			drawable: theDrawable,
			virtualHandVertexArrays: theVirtualHandVertexArrays,
			commandBuffer: theCommandBuffer)
		theDrawable.encodePresent(commandBuffer: theCommandBuffer)
		theCommandBuffer.commit()
		
		theFrame.endSubmission()
	}

	func renderTheFrame(
		drawable: LayerRenderer.Drawable,
		virtualHandVertexArrays: [PlayersHand: [NEBVertex]],
		commandBuffer: MTLCommandBuffer
	) {

		let theRenderPassDescriptor = ggvrRenderPassDescriptor(
										layerRenderer: layerRenderer,
										drawable: drawable)

		let theViewProjection = ggvrMakeViewProjection(
			virtualHeadsetPose: virtualHeadsetPose ?? matrix_identity_double4x4,
			drawable: drawable,
			spaceGeometry: gameState.spaceGeometry,
			adjustPerceivedDistances: adjustPerceivedDistances)

		let theMessageUniforms = makeMessageUniforms(
									drawable: drawable,
									messageTexture: messageTexture)

		if let theRenderEncoder = commandBuffer.makeRenderCommandEncoder(
										descriptor: theRenderPassDescriptor) {
			theRenderEncoder.label = "Non-Euclidean Billiards render encoder"

			ggvrEncodeViewports(
				drawable: drawable,
				renderEncoder: theRenderEncoder)

			encodeCommands(
				renderEncoder: theRenderEncoder,
				viewProjection: theViewProjection,
				virtualHandVertexArrays: virtualHandVertexArrays,
				messageUniforms: theMessageUniforms)
				
			theRenderEncoder.endEncoding()
		}
	}
	
	func encodeCommands(
		renderEncoder: MTLRenderCommandEncoder,
		viewProjection: GGVRViewProjection,
		virtualHandVertexArrays: [PlayersHand: [NEBVertex]],
		messageUniforms: MessageUniforms?
	) {
		let theMatrixSize = MemoryLayout<simd_float4x4>.stride

		renderEncoder.pushDebugGroup("Draw the whole scene")

		renderEncoder.setCullMode(.back)			//	our default
		renderEncoder.setFrontFacing(.clockwise)	//	Metal's default (and ours)

		var theViewProjection = viewProjection	//	setVertexBytes requires a 'var'
		renderEncoder.setVertexBytes(
						&theViewProjection,
						length: MemoryLayout<GGVRViewProjection>.stride,
						index: BufferIndexViewProjection)
		
		//	Table
		//
		//	The table must be rendered first,
		//	so the grounding shadows and the pockets
		//	can be rendered on top of it.

		renderEncoder.setRenderPipelineState(billiardsTablePipelineState)
		renderEncoder.setDepthStencilState(enabledDepthStencilState)
		renderEncoder.setVertexBuffer(
						tableMesh.vertexBuffer,
						offset: 0,
						index: BufferIndexVertexAttributes)

		var theTablePose = matrix_identity_float4x4	//	table pose in table coordinates
		renderEncoder.setVertexBytes(
			&theTablePose,
			length: theMatrixSize,
			index: BufferIndexModelMatrix)

		renderEncoder.drawIndexedPrimitives(
						type: .triangle,
						indexCount: tableMesh.indexCount,
						indexType: .uint16,
						indexBuffer: tableMesh.indexBuffer,
						indexBufferOffset: 0,
						instanceCount: 1)
		
		//	Grounding shadows
		//
		//	The grounding shadows must get rendered directly on top of the table.
		//	The groundingShadowDepthStencilState neither reads from nor writes to
		//	the depth buffer.
		
		renderEncoder.setRenderPipelineState(groundingShadowPipelineState)
		renderEncoder.setDepthStencilState(groundingShadowDepthStencilState)
		renderEncoder.setVertexBuffer(
						groundingShadowMesh.vertexBuffer,
						offset: 0,
						index: BufferIndexVertexAttributes)

		for i in 0 ..< numBalls {
		
			let theBall = gameState.balls[i]

			if theBall.status == .inPlay {
		
				var theShadowPose = convertDouble4x4toFloat4x4(theBall.placement)
				renderEncoder.setVertexBytes(
					&theShadowPose,
					length: theMatrixSize,
					index: BufferIndexModelMatrix)

				renderEncoder.drawIndexedPrimitives(
								type: .triangle,
								indexCount: groundingShadowMesh.indexCount,
								indexType: .uint16,
								indexBuffer: groundingShadowMesh.indexBuffer,
								indexBufferOffset: 0,
								instanceCount: 1)
			}
		}
		
		//	Pockets
		//
		//	The pockets must get rendered immediately
		//	after the table and the grounding shadows.
		//	The pocketPipelineState enables clearDepthBuffer,
		//	so in effect it erases any and all preexisting
		//	content where we're drawing the pockets.
		
		renderEncoder.setRenderPipelineState(pocketPipelineState)
		renderEncoder.setDepthStencilState(pocketDepthStencilState)
		renderEncoder.setVertexBuffer(
						pocketMesh.vertexBuffer,
						offset: 0,
						index: BufferIndexVertexAttributes)

		for thePose in gameState.spaceGeometry.pocketPoses {
		
			var thePocketPose = convertDouble4x4toFloat4x4(thePose)
			renderEncoder.setVertexBytes(
				&thePocketPose,
				length: theMatrixSize,
				index: BufferIndexModelMatrix)

			renderEncoder.drawIndexedPrimitives(
							type: .triangle,
							indexCount: pocketMesh.indexCount,
							indexType: .uint16,
							indexBuffer: pocketMesh.indexBuffer,
							indexBufferOffset: 0,
							instanceCount: 1)
		}

		//	Now that we've drawn the table, grounding shadows and pockets
		//	(which use the depth buffer in unusual ways),
		//	we can draw all the "normal" content.

		//	Balls
		
		let theWorldUpVector = SIMD4<Double>(0.0, 1.0, 0.0, 0.0)

		renderEncoder.setRenderPipelineState(billiardBallPipelineState)
		renderEncoder.setDepthStencilState(enabledDepthStencilState)
		renderEncoder.setVertexBuffer(
						ballMesh.vertexBuffer,
						offset: 0,
						index: BufferIndexVertexAttributes)
		renderEncoder.setFragmentSamplerState(
						ballSampler,
						index: SamplerIndexPrimary)
			
		for i in 0 ..< numBalls {

			let theBall = gameState.balls[i]

			if theBall.status == .inPlay {

				var theBallPose = convertDouble4x4toFloat4x4(
									theBall.placement	//	right-to-left matrix action
								  * gameState.spaceGeometry.ballLift
								  * theBall.rotation)
				renderEncoder.setVertexBytes(
					&theBallPose,
					length: theMatrixSize,
					index: BufferIndexModelMatrix)
				
				let theUpVector4 = geometricInverse(
									of: theBall.rotation,
									geometryType: gameState.spaceGeometry.geometryType)
								 * theWorldUpVector
				var theUpVector3 = SIMD3<Float16>(
										Float16(theUpVector4.x),
										Float16(theUpVector4.y),
										Float16(theUpVector4.z))
				renderEncoder.setVertexBytes(
					&theUpVector3,
					length: MemoryLayout<SIMD3<Float16>>.stride,
					index: BufferIndexUpVector)

				var theColorSaturation = Float16(
					makeScreenshots ?
						1.0 :						//	Color all balls
						theBall.colorSaturation)	//	Color active player's balls only
				renderEncoder.setFragmentBytes(
					&theColorSaturation,
					length: MemoryLayout<Float16>.stride,
					index: BufferIndexColoration)

				renderEncoder.setFragmentTexture(
								ballTextures[i],
								index: TextureIndexPrimary)

				renderEncoder.drawIndexedPrimitives(
								type: .triangle,
								indexCount: ballMesh.indexCount,
								indexType: .uint16,
								indexBuffer: ballMesh.indexBuffer,
								indexBufferOffset: 0,
								instanceCount: 1)
			}
		}

		if let theInHandCueBallPose = gameState.inHandCueBallPose {	//	in table coordinates
		
			var theInHandPose = convertDouble4x4toFloat4x4(theInHandCueBallPose)
			renderEncoder.setVertexBytes(
				&theInHandPose,
				length: theMatrixSize,
				index: BufferIndexModelMatrix)

			//	The inHandCueBallPose is a plain translation, so its
			//	local up-vector is a reasonable global up-vector.
			var theUpVector = SIMD3<Float16>(0.0, 1.0, 0.0)
			renderEncoder.setVertexBytes(
				&theUpVector,
				length: MemoryLayout<SIMD3<Float16>>.stride,
				index: BufferIndexUpVector)

			var theColorSaturation = Float16(1.0)
			renderEncoder.setFragmentBytes(
				&theColorSaturation,
				length: MemoryLayout<Float16>.stride,
				index: BufferIndexColoration)

			renderEncoder.setFragmentTexture(
							ballTextures[0],
							index: TextureIndexPrimary)

			renderEncoder.drawIndexedPrimitives(
							type: .triangle,
							indexCount: ballMesh.indexCount,
							indexType: .uint16,
							indexBuffer: ballMesh.indexBuffer,
							indexBufferOffset: 0,
							instanceCount: 1)
		}
		
		//	Cue sticks
		
		renderEncoder.setRenderPipelineState(cueStickPipelineState)
		renderEncoder.setDepthStencilState(enabledDepthStencilState)
		renderEncoder.setVertexBuffer(
						cueStickMesh.vertexBuffer,
						offset: 0,
						index: BufferIndexVertexAttributes)
		renderEncoder.setFragmentTexture(
						cueStickTexture,
						index: TextureIndexPrimary)
		renderEncoder.setFragmentSamplerState(
						cueStickSampler,
						index: SamplerIndexPrimary)

		func renderCueStick(
			_ cueStick: CueStick
		) {
			if let theCueStickPose = cueStick.current,
			   cueStick.opacity > 0.0 {

				var theCueStickPoseFloat = convertDouble4x4toFloat4x4(theCueStickPose)
				renderEncoder.setVertexBytes(
					&theCueStickPoseFloat,
					length: theMatrixSize,
					index: BufferIndexModelMatrix)

				var theOpacity = Float16(cueStick.opacity)
				renderEncoder.setFragmentBytes(
					&theOpacity,
					length: MemoryLayout<Float16>.stride,
					index: BufferIndexColoration)

				renderEncoder.drawIndexedPrimitives(
								type: .triangle,
								indexCount: cueStickMesh.indexCount,
								indexType: .uint16,
								indexBuffer: cueStickMesh.indexBuffer,
								indexBufferOffset: 0,
								instanceCount: 1)
			}
		}
		renderCueStick(gameState.cueStick[.human]!)
		renderCueStick(gameState.cueStick[.computer]!)
	
		//	Hands
			
		//	To keep the triangle count low, the player's hands
		//	are 2D surfaces, not 3D models.

		renderEncoder.setRenderPipelineState(handPipelineState)
		renderEncoder.setDepthStencilState(enabledDepthStencilState)

		for thePlayersHand in PlayersHand.allCases {

			guard let theVirtualHandVertexArray = virtualHandVertexArrays[thePlayersHand]
			else { continue }

			//	The hand vertices are defined in the virtual table space,
			//	so the hand pose is always the identity.
			var theIdentityMatrix = matrix_identity_float4x4
			renderEncoder.setVertexBytes(
				&theIdentityMatrix,
				length: theMatrixSize,
				index: BufferIndexModelMatrix)

			renderEncoder.setVertexBytes(
				theVirtualHandVertexArray,
				length: theVirtualHandVertexArray.count * MemoryLayout<NEBVertex>.stride,
				index: BufferIndexVertexAttributes)
			
			//	Render back of hand
			
			//	The left and right hand meshes are topologically
			//	mirror images of each other.
			switch thePlayersHand {
			case .left:  renderEncoder.setCullMode(.front)
			case .right: renderEncoder.setCullMode(.back)
			}

			//	A selection of reasonable skin tones
			//	in linear Display P3:
			//
			//		SIMD3<Float16>(0.2351, 0.0966, 0.0271)
			//		SIMD3<Float16>(0.5067, 0.2492, 0.0765)
			//		SIMD3<Float16>(0.6863, 0.4235, 0.1712)
			//		SIMD3<Float16>(0.8192, 0.5507, 0.2408)
			//		SIMD3<Float16>(0.9482, 0.7180, 0.4439)
			//
			var theBackColor = SIMD4<Float16>(0.59, 0.33, 0.12, 1.0)
			renderEncoder.setFragmentBytes(
				&theBackColor,
				length: MemoryLayout<SIMD4<Float16>>.stride,
				index: BufferIndexColoration)

			renderEncoder.drawIndexedPrimitives(
							type: .triangle,
							indexCount: gHandMeshFacets.count,
							indexType: .uint16,
							indexBuffer: handFacets,
							indexBufferOffset: 0)
			
			//	Render palm of hand
			
			//	Draw the faces that got culled when drawing
			//	the back of the hand (above).
			switch thePlayersHand {
			case .left:  renderEncoder.setCullMode(.back)
			case .right: renderEncoder.setCullMode(.front)
			}

			//	See skin tones above.
			var thePalmColor = SIMD4<Float16>(0.82, 0.55, 0.24, 1.0)
			renderEncoder.setFragmentBytes(
				&thePalmColor,
				length: MemoryLayout<SIMD4<Float16>>.stride,
				index: BufferIndexColoration)

			renderEncoder.drawIndexedPrimitives(
							type: .triangle,
							indexCount: gHandMeshFacets.count,
							indexType: .uint16,
							indexBuffer: handFacets,
							indexBufferOffset: 0)
		}
		renderEncoder.setCullMode(.back)	//	restore our default

		//	Checkerboard tiling

		renderEncoder.setRenderPipelineState(checkerboardTilingPipelineState)
		renderEncoder.setDepthStencilState(enabledDepthStencilState)
		renderEncoder.setCullMode(.none)

		//	The tiling matrices are defined in the virtual table space,
		//	so the model matrix is the identity.
		var theIdentityMatrix = matrix_identity_float4x4
		renderEncoder.setVertexBytes(
			&theIdentityMatrix,
			length: theMatrixSize,
			index: BufferIndexModelMatrix)

		renderEncoder.setVertexBuffer(
						checkerboardPoses,
						offset: 0,
						index: BufferIndexTilingMatrices)
		
		func setColoredWall(
			_ wall: CellWall
		) {
		
			renderEncoder.setVertexBuffer(
							wall.vertices,
							offset: 0,
							index: BufferIndexVertexAttributes)

			var theColor = wall.color
			renderEncoder.setFragmentBytes(
				&theColor,
				length: MemoryLayout<SIMD3<Float16>>.stride,
				index: BufferIndexColoration)
		}

		//	If checkerboardCellWallsReducedLOD is non-empty,
		//	we'll first draw the high LOD instances
		//	and then draw the low LOD instances.
		
		precondition(
			checkerboardPoses.length % MemoryLayout<simd_float4x4>.stride == 0,
			"checkerboardPoses.length is not an integer multiple of the simd_float4x4 stride." )

		let theTotalPoseCount
			= checkerboardPoses.length
			/ MemoryLayout<simd_float4x4>.stride

		let theHighLODPoseCount = checkerboardCellWallsReducedLOD.isEmpty ?
				theTotalPoseCount : min( 1 + 30, theTotalPoseCount )

		let theLowLODPoseCount = theTotalPoseCount - theHighLODPoseCount
		let theLowLODBaseInstance = theHighLODPoseCount

		for wall in checkerboardCellWalls {
		
			setColoredWall(wall)
			
			let theHighLODBaseInstance
				= wall.mayCoincideWithPoolTable ? 1 : 0
			let theHighLODInstanceCount
				= theHighLODPoseCount - theHighLODBaseInstance

			if theHighLODInstanceCount > 0 {
			
				renderEncoder.drawPrimitives(
								type: .triangleStrip,
								vertexStart: 0,
								vertexCount: wall.vertexCount,
								instanceCount: theHighLODInstanceCount,
								baseInstance: theHighLODBaseInstance)
			}
		}
		if theLowLODPoseCount > 0 {

			for wall in checkerboardCellWallsReducedLOD {

				setColoredWall(wall)

				renderEncoder.drawPrimitives(
								type: .triangleStrip,
								vertexStart: 0,
								vertexCount: wall.vertexCount,
								instanceCount: theLowLODPoseCount,
								baseInstance: theLowLODBaseInstance)
			}
		}

		renderEncoder.setCullMode(.back)	//	restore our default

		//	Euclidean headset-based message
		//
		//	The message (if any) must get rendered last,
		//	to draw the message on top of all other content,
		//	regardless of depth.
		//
		//		Design note: It would be difficult to insert
		//		the message between other content because
		//		Pool Hall's GPU functions often use radial depth
		//		instead of the traditional planar depth.
		//
		
		if let theMessageTexture = messageTexture,
		   var theMessageUniforms = messageUniforms,
		   var theMessageOpacity = messageOpacity {
		   
			renderEncoder.setVertexBytes(
							&theMessageUniforms,
							length: MemoryLayout<MessageUniforms>.stride,
							index: MessageBufferIndexUniforms)
			renderEncoder.setRenderPipelineState(messagePipelineState)
			renderEncoder.setDepthStencilState(messageDepthStencilState)
			renderEncoder.setVertexBuffer(
							messageMesh,
							offset: 0,
							index: MessageBufferIndexVertexAttributes)
			renderEncoder.setFragmentBytes(
							&theMessageOpacity,
							length: MemoryLayout<SIMD3<Float16>>.stride,
							index: BufferIndexOpacity)
			renderEncoder.setFragmentTexture(
							theMessageTexture,
							index: MessageTextureIndex)
			renderEncoder.setFragmentSamplerState(
							messageSampler,
							index: MessageSamplerIndex)
			renderEncoder.drawPrimitives(
							type: .triangleStrip,
							vertexStart: 0,
							vertexCount: 4)
		}

		renderEncoder.popDebugGroup()
	}
}
